#include <tools/eigen_wrap.hpp>
#include <adelie_core/glm/glm_binomial.ipp>

template class adelie_core::glm::GlmBinomialLogit<float>;
template class adelie_core::glm::GlmBinomialLogit<double>;
template class adelie_core::glm::GlmBinomialProbit<float>;
template class adelie_core::glm::GlmBinomialProbit<double>;